#!/usr/bin/python
program_name = "printconf-gui"
program_authors = ["Crutcher Dunnavant", "The Fifth Royal Canadian Termite Brigade"]
## Copyright (C) 2001-2002 Red Hat, Inc.
## Copyright (C) 2001 Crutcher Dunnavant <crutcher@redhat.com>
## Copyright (C) 2002 Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


# =====================================================================================================================
# Style
# -----
# This program was written in vim, using python highlighting, and 132 collumn display.
# If a 132 column display is unavailable, please return to the 1970s, with your hardware.


# =====================================================================================================================
# Purpose
# -------
# This program provides the grahical mode components to printconfs front-end.

# =====================================================================================================================
# printconf
# ---------
# Tie ourself to the main lib

# import the general parts of printconf
from printconf_conf import *

# python mangles '_' for us, as "private" data. Whatever, its dumb. We re-setup the '_' function alias here, of the
# gettext module we imported from printconf_conf
_=gettext.gettext

program_copyright = _("Red Hat, Inc")
program_comments = _("\"The Trees! The Trees!\"  -Thoreau") + "\n\n" + \
		_("Thanks to everyone involved at http://www.linuxprinting.org!")

# Program Data Paths
# ------------------
# printconf-gui.glade file, which specifies the user interface layout, is available at:
glade_file = conf.printconf_gui_dir + "/printconf-gui.glade"
if os.environ.has_key("PERLLIB"):
	del os.environ["PERLLIB"]

if os.environ.has_key("PERL5LIB"):
	del os.environ["PERL5LIB"]

# =====================================================================================================================
# Main NameSpace
# --------------

main = NameSpace()

# =====================================================================================================================
# Libs
# ----

# System Signals
# --------------
# 
# Since this program runs as a graphical interface, we really want to ignore whatever is happening on the console.
# Importing the signal module makes it possible to register the default handler (ignore) on terminal interupts.
import signal
debug_print("registering signal.SIG_DFL to handle signal.SIGINT")
signal.signal (signal.SIGINT, signal.SIG_DFL)


# Glade Display Initialization
# ----------------------------
# 
# This program uses gnome and gtk for its widget set, and glade as the layout template for the interface. Since there
# is a good deal of initialization to be done before the program is active, we go ahead and set up the basics of this
# interface here, and get it up on the screen before initalizing the rest of the program. First, import the modules
# necesary for displaying the interface:
os.environ["PYGTK_FATAL_EXCEPTIONS"] = "1"
import gtk, gnome.ui, libglade, gnome.xmhtml

def spin_gtk():
	while gtk.events_pending():
		gtk.mainiteration()

# now, initalize an instance of the interfaces discribed in the glade context.
main.glade_context = libglade.GladeXML(glade_file, domain="printconf")

def widg(name):
	return main.glade_context.get_widget(name)

import GDK
busyCursor = gtk.cursor_new(GDK.WATCH)
pointerCursor = gtk.cursor_new(GDK.LEFT_PTR)

# =====================================================================================================================
# Gtk Signal Handler Registration
# -------------------------------
# 
# gtk is a callback based widget system, and so functions must be wired up to callbacks. Instead of collecting
# all such function references at the registration point, I prefer to incrementally collect handler functions and
# signal names as the handlers are defined; because I feel that the code is more readable that way.
main.gtk_signal_handlers = {}



# =====================================================================================================================
# Gnome Help
# ----------
#
# Gnome (and every other widget set) provide hooks to launch web pages. For some bizare reason, Gnome does this though
# a module named "help". They claim it is because it was written to launch their help pages, but it doesnt matter, it
# can be used to spawn off arbitrary web content. We do, however, us it to launch our help pages through out. It
# requires the following modlue:
import gnome.help

# =====================================================================================================================
# Generic Dialogs
# ---------------
#
# At frequent intervals in the code, the necessity of presenting an informational or inquisative dialog to the user
# presents itself. In these scenarios, it is often necessary to only show the user a single message, a status symbol,
# and then wait for a decisision to be made. The following functions provide small wrappers arround these actions.
# We use our own wrappers here in order to minimize the difference in code between the gui and tui versions.
def gui_error(message):
	called(gui_error)
        err_dialog = gnome.ui.GnomeErrorDialog(message,
					       widg("queue_list_dialog"))
        err_dialog.run_and_close()

def gui_warn(message): 
	called(gui_warn)
        warn_dialog = gnome.ui.GnomeWarningDialog(message,
						  widg("queue_list_dialog"))
        warn_dialog.run_and_close()

def gui_info(message):
	called(gui_info)
        info_dialog = gnome.ui.GnomeOkDialog(message,
					     widg("queue_list_dialog"))
        info_dialog.run_and_close()

def gui_ask(title, message):
	called(gui_ask)
	# FIXME use the title
        ask_dialog = gnome.ui.GnomeOkCancelDialog(message, lambda *x: None,
						  widg("queue_list_dialog"))
	# This equates to:
	# 	if and only if they click "Ok", then return true, else return false.
        return not ask_dialog.run_and_close()

def gui_request(message, prompt, default, validate_func):
	called(gui_request)

	dialog = gnome.ui.GnomeDialog("", gnome.uiconsts.STOCK_BUTTON_OK, gnome.uiconsts.STOCK_BUTTON_CANCEL)
	dialog.set_parent(widg("queue_list_dialog"))
	table = gtk.GtkTable(rows = 2, cols = 2)
	dialog.vbox.pack_start(table)

	table.set_row_spacings(4)
	table.set_col_spacings(4)

	m_label = gtk.GtkLabel(message)
	m_label.set_alignment(0.0, 0.5)
	table.attach(m_label, 0, 2, 0, 1)
	
	p_label = gtk.GtkLabel(prompt)
	p_label.set_alignment(1.0, 0.5)
	table.attach(p_label, 0, 1, 1, 2)

	entry = gtk.GtkEntry(20)
	entry.set_text(default)
	table.attach(entry, 1, 2, 1, 2)

	dialog.vbox.show_all()

	while 1:
		retval = dialog.run()
		if retval == 0:
			text = entry.get_text()
			if validate_func(text):
				dialog.hide()
				return text
		else:
			dialog.hide()
			return



# =====================================================================================================================
# Debug Window
# ------------

debug_window = NameSpace()

#
# A small window containing the labels "Foo", "Bar", "Baz", "Qux", and "Bob" is included in the glade context. It is
# meant as an anchor for debugging commands during development. This window gets displayed if debuging is turned on.
# Fetch the window from the glade context, and show it if neccessary.
debug_window.window = widg("debug_window")
if __debug__:
	debug_print("Showing debug window")
	debug_window.window.show()

# If debugging is turned on, the debugging interface should not go away. This handler registers a function to return 1
# on the gtk delete event for the debugging window, causing the event to be ignored.
main.gtk_signal_handlers["on_debug_window_delete_event"] = lambda *x: 1

# Initial debugging events are simple printouts of which event was called.
debug_window.functions = {	"foo" : lambda: debug_print("foo"),
				"bar" : lambda: debug_print("bar"),
				"baz" : lambda: debug_print("baz"),
				"qux" : lambda: debug_print("qux"),
				"bob" : lambda: debug_print("bob")}

# A generic function to make it easier to override debuging events dynamically.
def debug_window_button_handler(button, which):
	called(debug_window_button_handler)
	debug_window.functions[which]()

# Wire the debug handler up to the 5 generic functions.
main.gtk_signal_handlers.update( {	"on_debug_foo" : (debug_window_button_handler, "foo"),
					"on_debug_bar" : (debug_window_button_handler, "bar"),
					"on_debug_baz" : (debug_window_button_handler, "baz"),
					"on_debug_qux" : (debug_window_button_handler, "qux"),
					"on_debug_bob" : (debug_window_button_handler, "bob") } )

# =====================================================================================================================
# Exit Check Dialog
# ------------------

ecd = NameSpace()
ecd.title = _("Exit")
ecd.text = _("You have made changes, would you like to save them?\nIf you say no, your changes will be lost.")

def ecd_run():
	called(ecd_run)

	if conf.data_state == CURRENT:
		return 1

	dialog = gnome.ui.GnomeDialog(ecd.title,
			gnome.uiconsts.STOCK_BUTTON_YES,
			gnome.uiconsts.STOCK_BUTTON_NO,
			gnome.uiconsts.STOCK_BUTTON_CANCEL)

	label = gtk.GtkLabel(ecd.text)
	dialog.vbox.pack_start(label)
	label.show()

	retval = dialog.run_and_close()
	if retval == 0:
		# save and restart
		return srsd_run()
	elif retval == 1:
		# dont save or restart
		return 1
	else: # cancel
		return None


# =====================================================================================================================
# Save and Restart Status Dialog
# ------------------------------

def srsd_run():
	called(srsd_run)

	qld_busy()

	save_queues()
	if restart_lpd():
		gui_info(_("lpd restart succeeded."))
	else:
		gui_error(_("lpd restart failed."))

	qld_ready()

	return 1
	

# =====================================================================================================================
# New Queue Druid
# ---------------
nqd = NameSpace()
nqd.window = widg("nqd_window")
nqd.druid = widg("nqd_druid")

# The Start Page
# ---------------------
nqd.start = NameSpace()
nqd.start.page = widg("nqd_start_page")

# The NameType Page
# ---------------------
nqd.nametype = NameSpace()
nqd.nametype.page = widg("nqd_nametype_page")

nqd.nametype.bad_name = _("The name \"%s\" is bad. Please choose a name of the form [a-zA-Z][a-zA-Z0-9_-]*.")
nqd.nametype.taken_name = _("The name \"%s\" is already in use. Please choose a different name.")

# The TypeData Pages
# ---------------------
nqd.typedata = NameSpace()
nqd.typedata.local = NameSpace()
nqd.typedata.local.page = widg("nqd_typedata_local_page")
nqd.typedata.local.device_clist = widg("nqd_typedata_local_device_clist")

nqd.typedata.lpd = NameSpace()
nqd.typedata.lpd.page = widg("nqd_typedata_lpd_page")

nqd.typedata.smb = NameSpace()
nqd.typedata.smb.page = widg("nqd_typedata_smb_page")

nqd.typedata.ncp = NameSpace()
nqd.typedata.ncp.page = widg("nqd_typedata_ncp_page")

nqd.typedata.jetdirect = NameSpace()
nqd.typedata.jetdirect.page = widg("nqd_typedata_jetdirect_page")

nqd.typedata.pages = [nqd.typedata.local.page, nqd.typedata.lpd.page, nqd.typedata.smb.page,
			nqd.typedata.ncp.page, nqd.typedata.jetdirect.page]

nqd.typedata.typedict = {
	queue_types.local	: nqd.typedata.local,
	queue_types.lpd		: nqd.typedata.lpd,
	queue_types.smb		: nqd.typedata.smb,
	queue_types.ncp		: nqd.typedata.ncp,
	queue_types.jetdirect	: nqd.typedata.jetdirect
}

# The Driver Page
# ---------------------
nqd.driver = NameSpace()
nqd.driver.page = widg("nqd_driver_page")
nqd.driver.driver_ctree = widg("nqd_driver_driver_ctree")

# The Finish Page
# ---------------------
nqd.finish = NameSpace()
nqd.finish.page = widg("nqd_finish_page")

nqd.data = {}

def nqd_run():
	called(nqd_run)

	#conf.data_state = NOTSAVED
	nqd_init()
	nqd.window.show()


nqd.driver.node_dict = None
def nqd_init():
	called(nqd_init)

	rerender = 0
	locale = "C"
	if os.environ.has_key("LANG"):
		lang = os.environ["LANG"]
		if lang[0:2] == "zh" or lang[0:2] == "ko" or lang[0:2] == "ja":
			rerender = 1

		if lang[0:2] == "ja":
			locale = "ja_JP"

	nqd.data = {}
	flags = {"convert_text_to_Postscript" : 1,
		 "assume_data_is_text" : rerender,
		 "rerender_Postscript" : rerender}
	nqd.data["mf_flags"] = flags

	if locale != "C":
		nqd.data["filter_locale"] = locale

	nqd.driver_tuple = None
	nqd.queue_type_space = queue_types.local
	nqd.druid.set_page(nqd.start.page)

	if not nqd.driver.node_dict:
		nqd.driver.node_dict = driver_ctree(nqd.driver.driver_ctree)

	collapse_driver_ctree(nqd.driver.driver_ctree)
		
	# Calling push data with empty data will zero out the form
	nqd_push_data()


def nqd_pull_data():
	called(nqd_pull_data)

	def widg(name, context = main.glade_context):
		return context.get_widget(name)


	# Read the name and type from the nametype page
	nqd.data["queue_name"] = widg("nqd_nametype_name_textentry").get_text()
	if widg("nqd_nametype_typebuttons_local").get_active():
		nqd.queue_type_space = queue_types.local
	elif widg("nqd_nametype_typebuttons_lpd").get_active():
		nqd.queue_type_space = queue_types.lpd
	elif widg("nqd_nametype_typebuttons_smb").get_active():
		nqd.queue_type_space = queue_types.smb
	elif widg("nqd_nametype_typebuttons_ncp").get_active():
		nqd.queue_type_space = queue_types.ncp
	elif widg("nqd_nametype_typebuttons_jetdirect").get_active():
		nqd.queue_type_space = queue_types.jetdirect

	sel = nqd.typedata.local.device_clist.selection
	if sel:
		nqd.data["local_printer_device"] = nqd.typedata.local.device_clist.get_row_data(sel[0])

	nqd.data["lpd_server"] = widg("nqd_typedata_lpd_server_textentry").get_text()
	nqd.data["lpd_queue"] = widg("nqd_typedata_lpd_queue_textentry").get_text()
	nqd.data["lpd_strict_rfc1179"] = str(widg("nqd_typedata_lpd_strict_checkbutton").get_active())

	nqd.data["smb_share"] = widg("nqd_typedata_smb_share_textentry").get_text()
	nqd.data["smb_ip"] = widg("nqd_typedata_smb_ip_textentry").get_text()
	nqd.data["smb_workgroup"] = widg("nqd_typedata_smb_workgroup_textentry").get_text()
	nqd.data["smb_user"] = widg("nqd_typedata_smb_user_textentry").get_text()
	nqd.data["smb_password"] = widg("nqd_typedata_smb_password_textentry").get_text()
	nqd.data["smb_translate"] = str(widg("nqd_typedata_smb_translate_checkbutton").get_active())
	
	nqd.data["ncp_server"] = widg("nqd_typedata_ncp_server_textentry").get_text()
	nqd.data["ncp_queue"] = widg("nqd_typedata_ncp_queue_textentry").get_text()
	nqd.data["ncp_user"] = widg("nqd_typedata_ncp_user_textentry").get_text()
	nqd.data["ncp_password"] = widg("nqd_typedata_ncp_password_textentry").get_text()

	nqd.data["jetdirect_ip"] = widg("nqd_typedata_jetdirect_ip_textentry").get_text()
	nqd.data["jetdirect_port"] = widg("nqd_typedata_jetdirect_port_textentry").get_text()


def nqd_push_data():
	called(nqd_push_data)

	# write the name and type to the nametype page
	widg("nqd_nametype_name_textentry").set_text(nqd.data.get("queue_name", ""))
	if nqd.queue_type_space == queue_types.local:
		widg("nqd_nametype_typebuttons_local").set_active(1)
	if nqd.queue_type_space == queue_types.lpd:
		widg("nqd_nametype_typebuttons_lpd").set_active(1)
	if nqd.queue_type_space == queue_types.smb:
		widg("nqd_nametype_typebuttons_smb").set_active(1)
	if nqd.queue_type_space == queue_types.ncp:
		widg("nqd_nametype_typebuttons_ncp").set_active(1)
	if nqd.queue_type_space == queue_types.jetdirect:
		widg("nqd_nametype_typebuttons_jetdirect").set_active(1)

	widg("nqd_typedata_lpd_server_textentry").set_text(nqd.data.get("lpd_server", ""))
	widg("nqd_typedata_lpd_queue_textentry").set_text(nqd.data.get("lpd_queue", ""))
	widg("nqd_typedata_lpd_strict_checkbutton").set_active(int(nqd.data.get("lpd_strict_rfc1179", 0)))

	widg("nqd_typedata_smb_share_textentry").set_text(nqd.data.get("smb_share", ""))
	widg("nqd_typedata_smb_ip_textentry").set_text(nqd.data.get("smb_ip", ""))
	widg("nqd_typedata_smb_workgroup_textentry").set_text(nqd.data.get("smb_workgroup", ""))
	widg("nqd_typedata_smb_user_textentry").set_text(nqd.data.get("smb_user", ""))
	widg("nqd_typedata_smb_password_textentry").set_text(nqd.data.get("smb_password", ""))
	widg("nqd_typedata_smb_translate_checkbutton").set_active(int(nqd.data.get("smb_translate", 1)))
	
	widg("nqd_typedata_ncp_server_textentry").set_text(nqd.data.get("ncp_server", ""))
	widg("nqd_typedata_ncp_queue_textentry").set_text(nqd.data.get("ncp_queue", ""))
	widg("nqd_typedata_ncp_user_textentry").set_text(nqd.data.get("ncp_user", ""))
	widg("nqd_typedata_ncp_password_textentry").set_text(nqd.data.get("ncp_password", ""))

	widg("nqd_typedata_jetdirect_ip_textentry").set_text(nqd.data.get("jetdirect_ip", ""))
	widg("nqd_typedata_jetdirect_port_textentry").set_text(nqd.data.get("jetdirect_port", "9100"))


def nqd_handle_next(page, *args):
	called(nqd_handle_next)

	nqd_pull_data()

	# nametype -> typedata
	if page == nqd.nametype.page:
		queue_name = nqd.data["queue_name"]
		if not valid_queue_name(queue_name):
			gui_error(nqd.nametype.bad_name % queue_name)
			return 1

		if not check_queue_name_uniqueness(queue_name, qld.queue_dict_dict, qld.alias_dict_dict):
			gui_error(nqd.nametype.taken_name % queue_name)
			return 1

		if not nqd.queue_type_space.check() and \
		   not gui_ask(_("Warning"), nqd.queue_type_space.message):
			return 1

		nqd.druid.set_page(nqd.typedata.typedict[nqd.queue_type_space].page)
		if nqd.queue_type_space == queue_types.local:
			nqd_typedata_local_display_devices()

		return 1

	# typedata -> driver
	elif page in nqd.typedata.pages:
		if nqd.queue_type_space == queue_types.local:
			local_printer_device = nqd.data.get("local_printer_device")
			if not local_printer_device:
				gui_error(_("You must specify a device."))
				return 1

			dev = scan_local_printer_devices().get(local_printer_device)
			if dev and dev.has_key("printer"):
				printer = dev["printer"]

				printer_driver = printer.drivers[0]
				for driver in printer.drivers:
					black = driver_blacklist.dict.get(driver)
					if not black or black.check():
						printer_driver = driver
						break

				nqd.driver_tuple = (drivers.foomatic, (printer, printer_driver))

		elif nqd.queue_type_space == queue_types.lpd:
			if nqd.data["lpd_server"] == "":
				gui_error(_("You must specify a server."))
				return 1
			# This is legal, not overlooked
			#elif nqd.data["lpd_queue"] == "":
			#	gui_error(_("You must specify a queue."))
			#	return 1

		elif nqd.queue_type_space == queue_types.smb:
			if nqd.data["smb_share"] == "":
				gui_error(_("You must specify a SMB share to print to."))
				return 1

		elif nqd.queue_type_space == queue_types.ncp:
			if nqd.data["ncp_server"] == "":
				gui_error(_("You must specify an NCP server to print to."))
				return 1
			elif nqd.data["ncp_queue"] == "":
				gui_error(_("You must specify a queue on the NCP server to print to."))
				return 1

		elif nqd.queue_type_space == queue_types.jetdirect:
			if nqd.data["jetdirect_ip"] == "":
				gui_error(_("You must specify the ip address of a JetDirect printer."))
				return 1
			elif nqd.data["jetdirect_port"] == "":
				gui_error(_("You must specify the port of the JetDirect printer."))
				return 1

		nqd.druid.set_page(nqd.driver.page)
		nqd.druid.show_all()
		display_driver_ctree(nqd.driver.driver_ctree, nqd.driver.node_dict, nqd.driver_tuple)
		return 1

	# driver -> finish
	elif page == nqd.driver.page:
		sel = nqd.driver.driver_ctree.selection
		if not sel:
			gui_error(_("You must select a driver."))
			return 1

		driver_tuple = nqd.driver.driver_ctree.node_get_row_data(sel[0])
		
		if driver_tuple[0] == drivers.foomatic:
			black = driver_blacklist.dict.get(driver_tuple[1][1])
			if black and not black.check() and \
			   not gui_ask(_("Warning"), black.message):
					return 1

		nqd.driver_tuple = driver_tuple
		nqd.druid.set_page(nqd.finish.page)
		nqd_finish_describe_queue()
		return 1

	else:
		raise RuntimeError, "Unknown page"

	return 1


def nqd_handle_back(page, *args):
	called(nqd_handle_back)
	
	nqd_pull_data()

	# nametype -> start
	if page == nqd.nametype.page:
		# You cant go back from here.
		return 1

	# typedata -> nametype
	elif page in nqd.typedata.pages:
		nqd.druid.set_page(nqd.nametype.page)
		return 1

	# driver -> typedata
	elif page == nqd.driver.page:
		nqd.druid.set_page(nqd.typedata.typedict[nqd.queue_type_space].page)
		return 1

	# finish -> driver
	elif page == nqd.finish.page:
		nqd.druid.set_page(nqd.driver.page)
		return 1

	else:
		raise RuntimeError, "Unknown page"

	return 1


def nqd_handle_cancel(*args):
	called(nqd_handle_cancel)

	nqd.window.hide()

	qld_ready()

	return 1


def nqd_handle_finish(page, *args):
	called(nqd_handle_finish)

	if page == nqd.finish.page:
		if not nqd.queue_type_space.check() and \
		   not gui_ask(_("Warning"), nqd.queue_type_space.message):
			return 1

		if nqd.driver_tuple[0] == drivers.foomatic:
			black = driver_blacklist.dict.get(nqd.driver_tuple[1][1])
			if black and not black.check():
				if not gui_ask(_("Warning"), black.message):
					return

		construct_queue(nqd.queue_type_space, nqd.data, nqd.driver_tuple)
		qld.selected_queue_name = nqd.data["queue_name"]
		qld_list_queues()
		nqd.window.hide()

		qld_ready()
	else:
		raise RuntimeError, "Unknown page"


def nqd_typedata_local_rescan_devices(*args):
	called(nqd_typedata_local_rescan_devices)
	scan_local_printer_devices(force = 1)
	nqd_typedata_local_display_devices()


nqd.typedata.local.custom = NameSpace()
nqd.typedata.local.custom.prompt = _("Specify the device to use.")
def nqd_typedata_local_custom_device(*args):
	called(nqd_typedata_local_custom_device)

	def setfunc(text, dict = nqd.data, key = "local_printer_device"):
		if text:
			dict[key] = text
			nqd_typedata_local_display_devices()

	dialog = gnome.ui.GnomeRequestDialog(0, nqd.typedata.local.custom.prompt, "", 100, setfunc, None)
	dialog.run_and_close()


def nqd_typedata_local_display_devices():
	called(nqd_typedata_local_display_devices)

	display_local_device_clist(nqd.typedata.local.device_clist, nqd.data.get("local_printer_device"))


nqd.finish.overview_text = _("About to create the following queue:")
nqd.finish.type_text = _("Type:")
nqd.finish.driver_text = _("Driver:")
def nqd_finish_describe_queue():
	called(nqd_finish_describe_queue)

	if nqd.queue_type_space == queue_types.local:
		type_desc = _("Device:") + " " + nqd.data["local_printer_device"]

	elif nqd.queue_type_space == queue_types.lpd:
		type_desc = _("Server:") + " " + nqd.data["lpd_server"] + "\n" + \
				_("Queue:") + " " + nqd.data["lpd_queue"]

	elif nqd.queue_type_space == queue_types.smb:
		type_desc = _("Share:") + " " + nqd.data["smb_share"] + "\n" + \
				_("Host:") + " " + nqd.data["smb_ip"] + "\n" + \
				_("Workgroup:") + " " + nqd.data["smb_workgroup"] + "\n" + \
				_("User:") + " " + nqd.data["smb_user"]

	elif nqd.queue_type_space == queue_types.ncp:
		type_desc = _("Server:") + " " + nqd.data["ncp_server"] + "\n" + \
				_("Queue:") + " " + nqd.data["ncp_queue"] + "\n" + \
				_("User:") + " " + nqd.data["ncp_user"]

	elif nqd.queue_type_space == queue_types.jetdirect:
		type_desc = _("Ip:") + " " + nqd.data["jetdirect_ip"] + "\n" + \
				_("Port:") + " " + nqd.data["jetdirect_port"]

	else:
		raise RuntimeError, "unknown type"

	if nqd.driver_tuple[0] == drivers.foomatic:
		(printer, driver) = nqd.driver_tuple[1]
		driver_desc = "%s %s %s" % (printer.make, printer.model, driver)
	else:
		driver_desc = nqd.driver_tuple[0].label


	nqd.finish.page.set_text("%s\n\n%s\n%s\n%s\n\n%s\n%s" % (nqd.finish.overview_text, nqd.finish.type_text,
				nqd.queue_type_space.long_pretty_name, type_desc, nqd.finish.driver_text, driver_desc))


main.gtk_signal_handlers.update({
	"on_nqd_window_delete_event"		: nqd_handle_cancel,
	"on_nqd_druid_cancel"			: nqd_handle_cancel,

	"on_nqd_nametype_page_back"		: nqd_handle_back,
	"on_nqd_typedata_local_page_back"	: nqd_handle_back,
	"on_nqd_typedata_lpd_page_back"		: nqd_handle_back,
	"on_nqd_typedata_smb_page_back"		: nqd_handle_back,
	"on_nqd_typedata_ncp_page_back"		: nqd_handle_back,
	"on_nqd_typedata_jetdirect_page_back"	: nqd_handle_back,
	"on_nqd_driver_page_back"		: nqd_handle_back,
	"on_nqd_finish_page_back"		: nqd_handle_back,

	"on_nqd_nametype_page_next"		: nqd_handle_next,
	"on_nqd_typedata_local_page_next"	: nqd_handle_next,
	"on_nqd_typedata_lpd_page_next"		: nqd_handle_next,
	"on_nqd_typedata_smb_page_next"		: nqd_handle_next,
	"on_nqd_typedata_ncp_page_next"		: nqd_handle_next,
	"on_nqd_typedata_jetdirect_page_next"	: nqd_handle_next,
	"on_nqd_driver_page_next"		: nqd_handle_next,

	"on_nqd_finish_page_finish"		: nqd_handle_finish,

	"on_nqd_typedata_local_rescan_button_clicked"	: nqd_typedata_local_rescan_devices,
	"on_nqd_typedata_local_custom_button_clicked"	: nqd_typedata_local_custom_device,
})


# =====================================================================================================================

def driver_ctree(ctree):
	called(driver_ctree)

	ctree.clear()
	ctree.set_line_style(gtk.CTREE_LINES_NONE)
	ctree.set_expander_style(gtk.CTREE_EXPANDER_TRIANGLE)

	ctree.set_auto_sort(1)

	node_dict = {}

	makes = foomatic.make_model_dict_dict.keys()
	makes.sort()
	for make in makes:
		make_node = ctree.insert_node(None, None, [make], is_leaf = 0)
		ctree.node_set_selectable(make_node, 0)

		model_dict = foomatic.make_model_dict_dict[make]
		models = model_dict.keys()
		models.sort()
		for model in models:
			model_node = ctree.insert_node(make_node, None, [model], is_leaf = 0)
			ctree.node_set_selectable(model_node, 0)

			printer = model_dict[model]
			for driver in printer.drivers:
				driver_node = ctree.insert_node(model_node, None, [driver], is_leaf = 1)
				driver_tuple = (drivers.foomatic, (printer, driver))
				node_dict[driver_tuple] = driver_node
				ctree.node_set_row_data(driver_node, driver_tuple)
	ctree.set_auto_sort(0)

	last_node = ctree.node_nth(0)
	for driver in [drivers.raw, drivers.text, drivers.postscript]:
		node = ctree.insert_node(None, last_node, [driver.label], is_leaf = 1)
		driver_tuple = (driver,)
		node_dict[driver_tuple] = node
		ctree.node_set_row_data(node, driver_tuple)
		last_node = node

	return node_dict

def collapse_driver_ctree(ctree):
	called(collapse_driver_ctree)

	ctree.freeze()

	# Collapse the printer tree
	for node in ctree.selection:
		ctree.unselect(node)
	for node in ctree.base_nodes():
		ctree.collapse_recursive(node)

	ctree.thaw()
	ctree.show_all()

fukxor = None
def display_driver_ctree(ctree, node_dict, driver_tuple):
	called(display_driver_ctree)

	ctree.freeze()

	node = node_dict.get(driver_tuple)
	if node:
		pnode = node.parent
		while pnode:
			ctree.expand(pnode)
			pnode = pnode.parent

	else:
		# The default selection is postscript
		node = node_dict.get((drivers.postscript,))

	ctree.select(node)
	ctree.thaw()

	ctree.node_moveto(node, 0, 0.5, 0)


def display_local_device_clist(clist, local_printer_device = None):
	called(display_local_device_clist)

	local_devices = scan_local_printer_devices(force = 0)

	clist.freeze()
	clist.clear()

	keys = local_devices.keys()
	keys.sort()
	for key in keys:
		device = local_devices[key]

		if device.has_key("printer"):
			desc = "%s %s" % (device["printer"].make, device["printer"].model)
		elif device.has_key("auto"):
			desc = "%s %s" % (device["auto"]["manufacturer"], device["auto"]["model"])
		else:
			desc = ""
	
		row = clist.append([device["device"], desc])
		clist.set_row_data(row, key)
	
	if local_printer_device:
		if local_devices.has_key(local_printer_device):
			row = clist.find_row_from_data(local_printer_device)
			clist.select_row(row, 0)
		else:
			clist.insert(0, [local_printer_device, _("Custom")])
			clist.set_row_data(0, local_printer_device)
			clist.select_row(0, 0)

	clist.thaw()


# =====================================================================================================================
# Queue Edit Dialog
# -----------------
qed = NameSpace()

qed.dialog = widg("queue_edit_dialog")
qed.dialog.close_hides(1)
qed.notebook = widg("qed_notebook")

def qed_busy():
	gtk.GtkWindow.set_sensitive(qed.dialog, 0)
	spin_gtk()
	

def qed_ready():
	gtk.GtkWindow.set_sensitive(qed.dialog, 1)
	spin_gtk()

qed.namealias = NameSpace()
qed.namealias.index = 0
qed.typedata = NameSpace()
qed.typedata.index = 1
qed.driver = NameSpace()
qed.driver.index = 2
qed.driveroptions = NameSpace()
qed.driveroptions.index = 3

def qed_init(queue_dict):
	called(qed_init)

	qed.queue_dict = queue_dict
	qed.queue = qed.queue_dict["queue"]


def qed_run(queue_dict):
	called(qed_run)

	# This is cheap. We just save all of the old state, and restore it, if necessary.
	backup_ctx = queue_edit.dynamic_queue_ctx.copy()
	backup_state = conf.data_state
	backup_name = queue_dict["queue"].name

	qed_init(queue_dict)

	qed_namealias_push()
	qed_typedata_push()
	qed_driver_push()

	while 1:
		pressed = qed.dialog.run()
		if pressed == 0: # Ok
			qed_busy()
			if not (qed_namealias_check_and_pull() and \
				qed_typedata_check_and_pull() and \
				qed_driver_check_and_pull()):
				qed_ready()
				continue

			qed_ready()

			conf.data_state = NOTSAVED
			qld.selected_queue_name = qed.queue.name

			break

		elif pressed == 2: # Help
			npage = qed.notebook.get_current_page()
			if npage == 0:
				help_page = "printconf-edit.html#PRINTCONF-EDIT-NAMES"
			elif npage == 1:
				tpage = qed.typedata.notebook.get_current_page()
				if tpage == 0:
					help_page = "printconf-edit-queuetype.html"
				elif tpage == 1:
					help_page = 'printconf-edit-queuetype.html'
				elif tpage == 2:
					help_page = 'printconf-edit-queuetype.html'
				elif tpage == 3:
					help_page = 'printconf-edit-queuetype.html'
				elif tpage == 4:
					help_page = 'printconf-edit-queuetype.html'
			elif npage == 2:
				help_page = "printconf-edit-driver.html"
			elif npage == 3:
				help_page = "printconf-edit-driver-options.html"

			path = "%s/%s" % (conf.printconf_help_dir, help_page)
			debug_print("launching help at %s" % path)

			qed_busy()

			gnome.help.goto("file://" + path)

			qed_ready()

		else: # Cancel
			debug_print("cancel")

			queue_edit.dynamic_queue_ctx = backup_ctx
			conf.data_state = backup_state
			qld.selected_queue_name = backup_name

			break

	qed.dialog.hide()

def qed_notebook_switch_page(notebook, obj, pageindex):
	called(qed_notebook_switch_page)
	qed_busy()
	qed.pagehandlers[pageindex](pageindex)
	qed_ready()

main.gtk_signal_handlers["on_qed_notebook_switch_page"] = qed_notebook_switch_page

# Namealias page
# ----------------------------------------------------------------------
qed.namealias.page = widg("qed_namealias_page")
qed.namealias.name_entry = widg("qed_namealias_name_entry")
qed.namealias.alias_clist = widg("qed_namealias_alias_clist")

def qed_namealias_push():
	called(qed_namealias_push)

	qed.notebook.set_page(0)

	qed.namealias.name_entry.set_text(qed.queue.name)

	qed.namealias.alias_clist.clear()
	aliases = qed.queue["alias_list"][:]
	aliases.sort(lambda a,b: cmp(a.value, b.value))
	for alias in aliases:
		row = qed.namealias.alias_clist.append([alias.value])
		qed.namealias.alias_clist.set_row_data(row, alias)


qed.namealias.bad_name = _("The name \"%s\" is bad. Please choose a name of the form [a-zA-Z][a-zA-Z0-9_-]*.")
qed.namealias.taken_name = _("The name \"%s\" is already in use. Please choose a different name.")
def qed_namealias_check_and_pull():
	called(qed_namealias_check_and_pull)

	def grab():
		qed.notebook.set_page(qed.namealias.index)
		qed.namealias.name_entry.grab_focus()
		qed.namealias.name_entry.select_region(0, -1)

	name = qed.namealias.name_entry.get_text()
	if not valid_queue_name(name):
		grab()
		gui_error(qed.namealias.bad_name % name)
		return None

	if not check_queue_name_uniqueness(name, qld.queue_dict_dict, qld.alias_dict_dict, qed.queue):
		grab()
		gui_error(qed.namealias.taken_name % name)
		return None

	qed.queue.name = name
	return 1


def qed_namealias_selected_alias():
	called(qed_namealias_selected_alias)

	sel = qed.namealias.alias_clist.selection
	if len(sel) == 0:
		return None
	return qed.namealias.alias_clist.get_row_data(sel[0])

def qed_namealias_verify_alias(alias):
	called(qed_namealias_verify_alias)

	if not valid_queue_name(alias):
		gui_error(qed.namealias.bad_name % alias)
		return None

	if not check_queue_name_uniqueness(alias, qld.queue_dict_dict, qld.alias_dict_dict, qed.queue):
		gui_error(qed.namealias.taken_name % alias)
		return None

	return 1
	

qed.namealias.add_prompt = _("Alias")
qed.namealias.add_message = _("Please choose a name of the form [a-zA-Z][a-zA-Z0-9_-]*.")
def qed_namealias_add_handler(*args):
	called(qed_namealias_add_handler)

	if not qed_namealias_check_and_pull():
		return

	alias_name = gui_request(qed.namealias.add_message, qed.namealias.add_prompt,
				"", qed_namealias_verify_alias)
	if not alias_name:
		return

	qed.queue["alias_list"].addData(AdmStringType, "alias").value = alias_name

	rectify_aliases(qed.queue)
	qed_namealias_push()


def qed_namealias_edit_handler(*args):
	called(qed_namealias_edit_handler)

	if not qed_namealias_check_and_pull():
		return

	alias = qed_namealias_selected_alias()
	if not alias:
		gui_error(_("You must select an alias to edit."))
		return

	alias_name = gui_request(qed.namealias.add_message, qed.namealias.add_prompt,
				alias.value, qed_namealias_verify_alias)
	if not alias_name:
		return

	alias.value = alias_name

	rectify_aliases(qed.queue)
	qed_namealias_push()


def qed_namealias_delete_handler(*args):
	called(qed_namealias_delete_handler)

	if not qed_namealias_check_and_pull():
		return

	alias = qed_namealias_selected_alias()
	if not alias:
		gui_error(_("You must select an alias to delete."))
		return

	alias.unlink()
	rectify_aliases(qed.queue)

	qed_namealias_push()

main.gtk_signal_handlers.update({
	"on_qed_namealias_add_button_clicked" : qed_namealias_add_handler,
	"on_qed_namealias_edit_button_clicked" : qed_namealias_edit_handler,
	"on_qed_namealias_delete_button_clicked" : qed_namealias_delete_handler,
})

# Typedata page
# ----------------------------------------------------------------------
qed.typedata.page = widg("qed_typedata_page")
qed.typedata.notebook = widg("qed_typedata_notebook")
qed.typedata.menu = widg("qed_typedata_menu")

qed.typedata.local_device_clist = widg("qed_typedata_local_device_clist")

def qed_typedata_push():
	called(qed_typedata_push)


	# pull the type data out of the queue
	typespace_setup(qed.queue, qed.typedata)

	# populate the widgets
	def push_s(name):
		widg("qed_typedata_%s_entry" % name).set_text(qed.typedata.data.get(name, ""))

	def push_b(name):
		widg("qed_typedata_%s_checkbox" % name).set_active(int(qed.typedata.data.get(name, 0)))

	qed.typedata.local_device_clist.clear()

	push_s("lpd_server")
	push_s("lpd_queue")
	push_b("lpd_strict_rfc1179")

	push_s("smb_share")
	push_s("smb_ip")
	push_s("smb_workgroup")
	push_s("smb_user")
	push_s("smb_password")
	push_b("smb_translate")

	push_s("ncp_server")
	push_s("ncp_queue")
	push_s("ncp_user")
	push_s("ncp_password")

	push_s("jetdirect_ip")
	push_s("jetdirect_port")


def qed_typedata_show(page_index = 1):
	called(qed_typedata_show)

	index = queue_types.type_list.index(qed.typedata.queue_type_space)
	qed.typedata.menu.set_history(index)
	qed.typedata.notebook.set_page(index)

	if qed.typedata.queue_type_space == queue_types.local:
		local_printer_device = qed_typedata_local_pull_device()
		qed.typedata.local_device_clist.clear()
		display_local_device_clist(qed.typedata.local_device_clist, local_printer_device)


def qed_typedata_check_and_pull():
	called(qed_typedata_check_and_pull)

	def show_me():
		qed.notebook.set_page(1)

	def pull_s(name):
		val = widg("qed_typedata_%s_entry" % name).get_text()
		qed.typedata.data[name] = val
		return val

	def pull_b(name):
		val = widg("qed_typedata_%s_checkbox" % name).get_active()
		qed.typedata.data[name] = val
		return val

	if not qed.typedata.queue_type_space.check() and \
	   not gui_ask(_("Warning"), qed.typedata.queue_type_space.message):
		show_me()
		return 0

	# pull the widgets that apply. Validate them.
	if qed.typedata.queue_type_space == queue_types.local:
		local_printer_device = qed_typedata_local_pull_device()

		if not local_printer_device:
			show_me()
			gui_error(_("You must specify a device."))
			return 0

	elif qed.typedata.queue_type_space == queue_types.lpd:
		if not pull_s("lpd_server"):
			show_me()
			gui_error(_("You must specify a server."))
			return 0

		if not pull_s("lpd_queue"):
			# this is legal, not overlooked.
			# show_me()
			# gui_error(_("You must specify a queue."))
			pass

		pull_b("lpd_strict_rfc1179")

	elif qed.typedata.queue_type_space == queue_types.smb:
		if not pull_s("smb_share"):
			show_me()
			gui_error(_("You must specify a SMB share to print to."))
			return 0

		pull_s("smb_ip")
		pull_s("smb_workgroup")
		pull_s("smb_user")
		pull_s("smb_password")
		pull_b("smb_translate")

	elif qed.typedata.queue_type_space == queue_types.ncp:
		if not pull_s("ncp_server"):
			show_me()
			gui_error(_("You must specify an NCP server to print to."))
			return 0

		if not pull_s("ncp_queue"):
			show_me()
			gui_error(_("You must specify a queue on the NCP server to print to."))
			return 0

		pull_s("ncp_user")
		pull_s("ncp_password")

	elif qed.typedata.queue_type_space == queue_types.jetdirect:
		if not pull_s("jetdirect_ip"):
			gui_error(_("You must specify the ip address of a JetDirect printer."))
			return 0

		if not pull_s("jetdirect_port"):
			qed.data["jetdirect_port"] = "9100"

	# if the widget pull succeded, push the type data back into the queue
	typespace_apply(qed.queue, qed.typedata)
	return 1

def qed_typedata_menu_selection(menu):
	called(qed_typedata_menu_selection)

	qed.typedata.queue_type_space = queue_types.type_list[menu.children().index(menu.get_active())]
	qed_typedata_show()

# Gotta love broken glade.
qed.typedata.menu.get_menu().signal_connect('selection-done', qed_typedata_menu_selection)


def qed_typedata_local_rescan_button_handler(*args):
	called(qed_typedata_local_rescan_button_handler)

	scan_local_printer_devices(force = 1)
	qed_typedata_show()

def qed_typedata_local_custom_button_handler(*args):
	called(qed_typedata_local_custom_button_handler)

	def setfunc(text, dict = qed.typedata.data, key = "local_printer_device"):
		if text:
			dict[key] = text
			qed.typedata.local_device_clist.clear()
			qed_typedata_show()

	dialog = gnome.ui.GnomeRequestDialog(0, _("Specify the device to use."), "", 100, setfunc, None)
	dialog.run_and_close()


def qed_typedata_local_autoselect_button_handler(*args):
	called(qed_typedata_local_autoselect_button_handler)

	local_printer_device = qed_typedata_local_pull_device()
	device = scan_local_printer_devices().get(local_printer_device)

	if not local_printer_device:
		gui_error(_("You must specify a device."))

	elif device and device.has_key("printer"):
		printer = device["printer"]

		printer_driver = printer.drivers[0]
		for driver in printer.drivers:
			black = driver_blacklist.dict.get(driver)
			if not black or black.check():
				printer_driver = driver
				break

		qed.driver.driver_tuple = (drivers.foomatic, (printer, printer_driver))

		sel = qed.driver.ctree.selection
		if sel:
			qed.driver.ctree.unselect(sel[0])

		qed.notebook.set_page(2)

	else:
		gui_error(_("Can't autoselect for this printer device."))

def qed_typedata_local_pull_device():
	called(qed_typedata_local_pull_device)

	sel = qed.typedata.local_device_clist.selection
	if sel:
		local_printer_device = qed.typedata.local_device_clist.get_row_data(sel[0])
		qed.typedata.data["local_printer_device"] = local_printer_device
	else:
		local_printer_device = qed.typedata.data.get("local_printer_device")

	return local_printer_device
	

main.gtk_signal_handlers.update({
	"on_qed_typedata_local_rescan_button_clicked"	: qed_typedata_local_rescan_button_handler,
	"on_qed_typedata_local_custom_button_clicked"	: qed_typedata_local_custom_button_handler,
	"on_qed_typedata_local_autoselect_button_clicked": qed_typedata_local_autoselect_button_handler
})

# Driver and Driveroptions page
# ----------------------------------------------------------------------
qed.driver.options_page = widg("qed_driveroptions_page")
qed.driver.page = widg("qed_driver_page")
qed.driver.ctree = widg("qed_driver_ctree")
qed.driver.options = NameSpace()
qed.driver.options.table = widg("qed_driver_options_table")

qed.driver.node_dict = None
def qed_driver_push():
	called(qed_driver_push)

	if not qed.driver.node_dict:
		qed.driver.node_dict = driver_ctree(qed.driver.ctree)

	driverspace_setup(qed.queue, qed.driver)
	qed.driver.driver_tuple = qed_driver_build_driver_tuple(qed.queue)
	collapse_driver_ctree(qed.driver.ctree)
	display_driver_ctree(qed.driver.ctree, qed.driver.node_dict, qed.driver.driver_tuple)

	qed_driver_options_wipe()

def qed_driver_show(pageindex):
	called(qed_driver_show)

	sel = qed.driver.ctree.selection
	if sel:
		qed.driver.driver_tuple = qed.driver.ctree.node_get_row_data(sel[0])

		qed.driver.f_type = qed.driver.driver_tuple[0].filter_type
		if qed.driver.f_type == "MAGICFILTER":
			qed.driver.foomatic.mf_type = qed.driver.driver_tuple[0].mf_type
			if qed.driver.foomatic.mf_type == "MFOMATIC":
				qed.driver.foomatic.printer_id = qed.driver.driver_tuple[1][0].id
				qed.driver.foomatic.gs_driver = qed.driver.driver_tuple[1][1]

	if pageindex == 2:
		display_driver_ctree(qed.driver.ctree, qed.driver.node_dict, qed.driver.driver_tuple)
		return
	
	# Ick. We going to show the driver options, but first, we have to read them.
	qed_driver_options_read()
	qed_driver_options_write()
	

def qed_driver_options_wipe():
	called(qed_driver_options_wipe)

	for child in qed.driver.options.table.children():
		qed.driver.options.table.remove(child)
	qed.driver.options.table.resize(1,2)

	qed.driver.options.widget_list = []
	qed.driver.options.driver_tuple = None

def qed_driver_options_read():
	called(qed_driver_options_read)

	for (label_widget, data_widget, option_tuple) in qed.driver.options.widget_list:
		(opt_type, en_shortname, prettyname, dict, default, type_data) = option_tuple
		key = (en_shortname, opt_type)

		if opt_type == "bool":
			dict[key] = data_widget.get_active()

		elif opt_type == "int":
			dict[key] = data_widget.get_value_as_int()

		elif opt_type == "float":
			dict[key] = data_widget.get_value_as_float()

		elif opt_type == "enum":
			dict[key] = data_widget.list.get_selection()[0].get_data("value")


def qed_driver_options_write():
	called(qed_driver_options_write)

	# We are current, if the options driver_tuple matches the driver_tuple
	if qed.driver.options.driver_tuple == qed.driver.driver_tuple:
		debug_print("already displayed for %s" % repr(qed.driver.driver_tuple))
		return
	else:
		qed.driver.options.driver_tuple = qed.driver.driver_tuple
		debug_print("displaying for %s" % repr(qed.driver.driver_tuple))

	qed_driver_options_wipe()

	option_list = generate_option_list(qed.driver)
	
	qed.driver.options.widget_list = []
	for option_tuple in option_list:
		(opt_type, en_shortname, prettyname, dict, default, type_data) = option_tuple
		def_val = dict.get((en_shortname, opt_type), default)

		label_widget = gtk.GtkLabel(label = prettyname)
		label_widget.set_alignment(0.0,0.5)
		
		if opt_type == "bool":
			data_widget = gtk.GtkCheckButton()
			data_widget.set_active(int(def_val))

		elif opt_type == "int":
			(max, min) = type_data
			val = int(def_val)
			adjustment = gtk.GtkAdjustment(val, lower = int(min), upper = int(max),
					step_incr = 1, page_incr = 10)
			data_widget = gtk.GtkSpinButton(adjustment)
			data_widget.set_digits(0)

		elif opt_type == "float":
			(max, min) = type_data
			val = float(def_val)
			adjustment = gtk.GtkAdjustment(val, lower = float(min), upper = float(max),
					step_incr = 0.1, page_incr = 1)
			data_widget = gtk.GtkSpinButton(adjustment)
			data_widget.set_digits(1)

		elif opt_type == "enum":
			val_list = type_data
			data_widget = gtk.GtkCombo()
			data_widget.entry.set_editable(0)
			set = def_val
			for (val, val_label) in val_list:
				item = gtk.GtkListItem(label = val_label)
				item.set_data("value", val)
				data_widget.list.append_items([item])
				item.show()

				if val == set:
					item.select()

		else:
			raise RuntimeError, "unknown type %s" % opt_type

		qed.driver.options.widget_list.append(label_widget, data_widget, option_tuple)


	if not len(qed.driver.options.widget_list):
		label = gtk.GtkLabel(label = _("There are no options available for this driver."))
		qed.driver.options.table.resize(1, 2)
		qed.driver.options.table.attach(label,0,2,0,1,gtk.FILL)
	else:
		qed.driver.options.table.resize(len(qed.driver.options.widget_list), 2)
		for i in range(len(qed.driver.options.widget_list)):
			(label_widget, data_widget) = qed.driver.options.widget_list[i][:2]
			qed.driver.options.table.attach(label_widget,0,1,i,i+1,gtk.FILL)
			qed.driver.options.table.attach(data_widget,1,2,i,i+1)

	qed.driver.options.table.show_all()



def qed_driver_build_driver_tuple(queue):
	called(qed_driver_build_driver_tuple)

	f_type = queue["filter_type"].value
	f_data = queue["filter_data"]
	
	if f_type == "NONE":
		driver_tuple = (drivers.raw,)
	elif f_type == "MAGICFILTER":
		mf_type = f_data["mf_type"].value
		if mf_type == "TEXT":
			driver_tuple = (drivers.text,)
		elif mf_type == "POSTSCRIPT":
			driver_tuple = (drivers.postscript,)
		elif mf_type == "MFOMATIC":
			printer_id = f_data["printer_id"].value
			gs_driver = f_data["gs_driver"].value

			foo_printer = foomatic.id_dict.get(printer_id)
			if not foo_printer:
				driver_tuple = (drivers.postscript,)
			else:
				if gs_driver in foo_printer.drivers:
					driver_tuple = (drivers.foomatic, (foo_printer, gs_driver))
				else:
					# Well, they list a driver we dont have. Punt
					driver_tuple = (drivers.foomatic, (foo_printer, foo_printer[0]))
	else:
		raise RuntimeError, "unknown type %s" % f_type

	return driver_tuple


def qed_driver_check_and_pull():
	called(qed_driver_check_and_pull)

	qed_driver_options_read()

	sel = qed.driver.ctree.selection
	if sel:
		qed.driver.driver_tuple = qed.driver.ctree.node_get_row_data(sel[0])

	if qed.driver.driver_tuple[0] == drivers.foomatic:
		black = driver_blacklist.dict.get(qed.driver.driver_tuple[1][1])
		if black and not black.check():
			if not gui_ask(_("Warning"), black.message):
				qed.notebook.set_page(qed.driver.index)
				return 0

	if qed.driver.driver_tuple[0] == drivers.raw:
		qed.driver.f_type = "NONE"
	elif qed.driver.driver_tuple[0] == drivers.text:
		qed.driver.f_type = "MAGICFILTER"
		qed.driver.foomatic.mf_type = "TEXT"
	elif qed.driver.driver_tuple[0] == drivers.postscript:
		qed.driver.f_type = "MAGICFILTER"
		qed.driver.foomatic.mf_type = "POSTSCRIPT"
	elif qed.driver.driver_tuple[0] == drivers.foomatic:
		qed.driver.f_type = "MAGICFILTER"
		qed.driver.foomatic.mf_type = "MFOMATIC"
		qed.driver.foomatic.printer_id = qed.driver.driver_tuple[1][0].id
		qed.driver.foomatic.gs_driver = qed.driver.driver_tuple[1][1]

	driverspace_apply(qed.queue, qed.driver)

	return 1
		
def qed_driver_ctree_show_handler(*args):
	called(qed_driver_ctree_show_handler)

	sel = qed.driver.ctree.selection
	if sel:
		qed.driver.ctree.node_moveto(sel[0], 0, 0.5, 0)

def qed_driver_notes_button_activate(html, info):
	called(qed_driver_notes_button_activate)

	if info.href:
		gnome.help.goto(info.href)

def qed_driver_notes_button_clicked(*args):
	called(qed_driver_notes_button_clicked)

	driver_tuple = None
	sel = qed.driver.ctree.selection
	if sel:
		driver_tuple = qed.driver.ctree.node_get_row_data(sel[0])

	source = ""
	message = _("Sorry,\nthere are no notes available for this printer.")
	if driver_tuple and driver_tuple[0] == drivers.foomatic:
		message = _("Notes from the Linux Printing Database\nhttp://www.linuxprinting.org\n\n")
		foo_printer_driver = foomatic_printer_driver_lookup(driver_tuple[1][0].id, driver_tuple[1][1])
		if foo_printer_driver:
			source = """
<HTML><BODY BGCOLOR=white>
<a href="http://www.linuxprinting.org">http://www.linuxprinting.org</a><br>
<h1>Printer Notes:</h1>
%s

<br><h1>Driver Notes:</h1>
%s
</BODY></HTML>""" % (foo_printer_driver.printer_comments_dict.get("en", ""),
			foo_printer_driver.printer_comments_dict.get("en", ""))

	dialog = gnome.ui.GnomeDialog("", gnome.uiconsts.STOCK_BUTTON_OK)

	message_label = gtk.GtkLabel(message)
	message_label.set_alignment(0.0, 0.5)
	dialog.vbox.pack_start(message_label)

	html = gnome.xmhtml.GtkXmHTML()
	html.set_usize(600, 400)
	html.set_allow_body_colors(gtk.TRUE)   
	html.connect('activate', qed_driver_notes_button_activate)

	# Piece of shit xmhtml!
	html.source("<HTML><BODY BGCOLOR=white></BODY></HTML>")
	html.source(source)
	dialog.vbox.pack_start(html)

	dialog.vbox.show_all()
	dialog.run_and_close()

main.gtk_signal_handlers.update({
	"on_qed_driver_ctree_show"		: qed_driver_ctree_show_handler,
	"on_qed_driver_notes_button_clicked"	: qed_driver_notes_button_clicked
})

# Fill in the page handlers
qed.pagehandlers = {
	0 : lambda x: None,
	1 : qed_typedata_show,
	2 : qed_driver_show,
	3 : qed_driver_show,
}

# =====================================================================================================================
# Queue List Dialog
# -----------------
qld = NameSpace()
qld.dialog = widg("queue_list_dialog")

def qld_busy():
	qld.dialog.get_window().set_cursor(busyCursor)
	qld.dialog.set_sensitive(0)
	spin_gtk()

def qld_ready():
	qld.dialog.get_window().set_cursor(pointerCursor)
	qld.dialog.set_sensitive(1)
	spin_gtk()


# Set up the queue clist
# ----------------------
qld.queue_clist = widg("qld_queue_clist")
qld.queue_clist.set_sort_column(1)
qld.queue_clist.set_auto_sort(1)
qld.queue_clist.set_column_width (0, 30)
qld.queue_clist.set_column_resizeable (0, 0)

# Status Images
# -------------
# 
# A small collection of status images is used to indicate queue status, and so now we load the GdkImlib module to read
# them in and make them available to gtk. The pixmaps are stored in the global dictionary "images", with descriptive
# names.
import GdkImlib
qld.images = {}
for (file, name) in [	("printer-D.png", "printer-default"),
			("printer-DS.png", "printer-default-static"),
			("printer-DO.png", "printer-default-overridden"),
			("printer-S.png", "printer-static"),
			("printer-O.png", "printer-overridden")]:
	im = GdkImlib.Image ("%s/%s" % (conf.printconf_gui_dir, file))
	im.render()
	qld.images[name] = im.get_pixmap()


qld.selected_queue_name = None
def qld_list_queues():

	# Look up the default queue name, so we can display it
	default_queue_name = get_default_queue_name()

	# refresh the queue dict dict
	(qld.queue_dict_dict, qld.alias_dict_dict) = get_queues()

	debug_print("early winter")
	# freeze the gtk widget that we are about to ravish, so as to reduce flicker
	qld.queue_clist.freeze()

	# throw out the old
	qld.queue_clist.clear()

	for queue_dict in sort_queues(qld.queue_dict_dict):

		queue = queue_dict["queue"]

		if not queue_dict["valid"]:
			row = qld.queue_clist.append(["", queue.name, "", "INVALID", _("Invalid Printer Record.")])
			qld.queue_clist.set_row_data(row, queue_dict)
			continue

		q_aliases = alias_list_string(queue)
		q_type = queue["queue_type"].value
		q_details = queue_details(queue)

		row = qld.queue_clist.append(["", queue.name, q_aliases, q_type, q_details])
		qld.queue_clist.set_row_data(row, queue_dict)

		if queue.name == default_queue_name:
			if not queue_dict["editable"]:
				(image, mask) = qld.images["printer-default-static"]
			elif queue_dict["override"]:
				(image, mask) = qld.images["printer-default-overridden"]
			else:
				(image, mask) = qld.images["printer-default"]

		else:
			if not queue_dict["editable"]:
				(image, mask) = qld.images["printer-static"]
			elif queue_dict["override"]:
				(image, mask) = qld.images["printer-overridden"]
			else:
				(image, mask) = (None, None)

		if image != None:
			qld.queue_clist.set_pixmap(row, 0, image, mask)
			

		if queue.name == qld.selected_queue_name:
			qld.queue_clist.select_row(row,0)
			qld.queue_clist.moveto(row, 0, 0.5, 0)

			
	# only use this once
	qld.selected_queue_name = None

	debug_print("looks like spring")
	# unfreeze the widget, and let the user gaze upon it in wonder!
	qld.queue_clist.thaw()


def qld_selected_queue_dict():
	called(qld_selected_queue_dict)

	sel = qld.queue_clist.selection
	if len(sel) == 0:
		return None
	return qld.queue_clist.get_row_data(sel[0])

	
#----------------------------------------------------------------
# the qld file menu
#

# Handler for the "Save Changes" option
def qld_on_save_activate(*args):
	called(qld_on_save_activate)

	save_queues()

# Handler for the "Restart Lpd" option
def qld_on_restart_lpd_activate(*args):
	called(qld_on_restart_lpd_activate)

	if restart_lpd():
		gui_info(_("lpd restart succeeded."))
	else:
		gui_error(_("lpd restart failed."))

# Handler for the "Override Queue" option
def qld_on_override_activate(*args):
	called(qld_on_override_activate)

	queue_dict = qld_selected_queue_dict()
	if not queue_dict:
		gui_error(_("You must select a printer queue to override."))
		return

	if queue_dict["editable"]:
		gui_error(_("You cannot override a locally defined printer."))
		return
		
	name = queue_dict["queue"].name
	override_queue(name)

	qld.selected_queue_name = name
	qld_list_queues()

# Handler for the "Exit" option
def qld_on_exit_activate(*args):
	called(qld_on_exit_activate)

	if ecd_run():
		gtk.mainquit()

	# Returning "1" stalls the delete event :)
	return 1

# Register the file menu option handlers
main.gtk_signal_handlers.update( {	"qld_on_save_activate" : qld_on_save_activate,
				"qld_on_restart_lpd_activate" : qld_on_restart_lpd_activate,
				"qld_on_override_activate" : qld_on_override_activate,
				"qld_on_exit_activate" : qld_on_exit_activate })

##----------------------------------------------------------------
## the qld test menu
##

tcd = NameSpace()
tcd.title = _("Test")
tcd.text = _("You have made changes, would you like to save them?\nIf you say no, you will not be able to print test pages.")
def test_check_dialog_run():
	called(test_check_dialog_run)

	if conf.data_state == CURRENT:
		return 1

	dialog = gnome.ui.GnomeDialog(tcd.title,
			gnome.uiconsts.STOCK_BUTTON_YES,
			gnome.uiconsts.STOCK_BUTTON_NO)

	label = gtk.GtkLabel(tcd.text)
	dialog.vbox.pack_start(label)
	label.show()

	retval = dialog.run_and_close()
	if retval == 0:
		# Yes
		return srsd_run()
	else:
		# No
		return None

def qld_build_test_menu():
	called(qld_build_test_menu)

	qld.print_tests = collect_print_tests()

	test_menu = gtk.GtkMenu()
	for (test_file, test_desc_dict) in qld.print_tests:
		item = gtk.GtkMenuItem(label = test_desc_dict["en"])
		item.signal_connect('activate', qld_test_handler, (test_file, test_desc_dict))
		test_menu.append(item)

	test_menu.show_all()
	widg("qld_menubar_test").set_submenu(test_menu)

def qld_test_handler(item, (test_file, test_desc_dict)):
	called(qld_test_handler)

	queue_dict = qld_selected_queue_dict()
        if not queue_dict:
                gui_error(_("You must select a printer queue to send a test page to."))
                return

	qld_busy()
	if not test_check_dialog_run():
		qld_ready()
		return

	queue_name = queue_dict["queue"].name
	if print_test_page(queue_name, conf.printconf_tests_dir + '/' + test_file):
		gui_info(_("Sent %s to \"%s\".") % (test_desc_dict["en"], queue_name))
	else:
		gui_error(_("There was an error trying to print the test page."))

	qld_ready()


# Register the test menu handlers
def qld_print_1337_ps_test_page_handler(*args):
	called(qld_print_1337_ps_test_page_handler)

	queue_dict = qld_selected_queue_dict()
        if not queue_dict:
                gui_error("j00 mU57 s3l3c7 4 Pr1|\\|73r 2 z3nD 4 t357 p3g3 2.")
                return

	qld_busy()
	if not test_check_dialog_run():
		qld_ready()
		return

	if print_1337_ps_test_page(queue_dict["queue"].name):
		gui_info("1337 p4g3 z3n7 2 \"%s\"." % (queue_dict["queue"].name))
	else:
		gui_error("sUc]<, j00 Pr1|\\|73r f41l3D!")
	qld_ready()

# Register the test menu handlers
main.gtk_signal_handlers.update({
	"on_qld_print_1337_ps_test_page_activate" : qld_print_1337_ps_test_page_handler,
})

#----------------------------------------------------------------
# The qld help menu
#

# Handler for the "About" option
def show_about(*args):
	called(show_about)

	about = gnome.ui.GnomeAbout(	title = "%s %s" % (program_name, printconf_version.version),
					list = program_authors,
					copyright = program_copyright,
					comments = program_comments)
	about.set_position(gtk.WIN_POS_CENTER)
	about.run()

def show_help(*args):
	called(show_help)
	
	qld_busy()

	gnome.help.goto("file://%s/index.html" %
			conf.printconf_help_dir)

	qld_ready()

main.gtk_signal_handlers["qld_on_about_activate"] = show_about

main.gtk_signal_handlers["qld_on_browse_help_activate"] = show_help

#----------------------------------------------------------------
# The qld menu bar
#

# Handler for the "New" Button on the qld menu bar
def qld_on_new_button_clicked(button, *args):
	called(qld_on_new_button_clicked)

	qld_busy()
	nqd_run()


# Handler for the "Edit" Button on the qld menu bar
def qld_on_edit_button_clicked(button, *args):
	called(qld_on_edit_button_clicked)

	queue_dict = qld_selected_queue_dict()
	if not queue_dict:
		gui_error(_("You must select a printer queue to edit."))
		return

	if not queue_dict["valid"]:
		gui_error(_("The printer's format is not understood, and can not be edited."))
		return

	if not queue_dict["editable"]:
		if gui_ask(_("Override?"), _("This is an imported printer. Do you want to create a local override?")):
			queue_name = queue_dict["queue"].name
			override_queue(queue_name)
			qld.selected_queue_name = queue_name
			qld_list_queues() 
			queue_dict = qld_selected_queue_dict()
		else:
			return

	qld_busy()

	qed_run(queue_dict)
	qld_list_queues()

	qld_ready()


# Handler for the "Delete" Button on the qld menu bar
def qld_on_delete_button_clicked(button, *args):
	called(qld_on_delete_button_clicked)

	queue_dict = qld_selected_queue_dict()
	if not queue_dict:
		gui_error(_("You must select a printer queue to delete."))
		return

	if not queue_dict["editable"]:
		gui_error(_("You cannot delete an imported printer."))
		return

	# This takes care of us if the queue is an override queue
	if queue_dict["override"]:
		qld.selected_queue_name = queue_dict["queue"].name

	delete_queue_and_fix_default(queue_dict["queue"])

	qld_list_queues()

# Handler for the "Default" Button on the qld menu bar
def qld_on_default_button_clicked(button, *args):
	called(qld_on_default_button_clicked)

	queue_dict = qld_selected_queue_dict()
	if not queue_dict:
		gui_error(_("You must select a printer queue\n to make the default."))
		return

	name = queue_dict["queue"].name

	# this makes the queue list dirty
	set_default_queue_name(name)

	qld.selected_queue_name = name
	qld_list_queues()

def qld_on_apply_button_clicked(button, *args):
	called(qld_on_apply_button_clicked)

	srsd_run()
	

# Register the qld menu bar buttons to be handled
main.gtk_signal_handlers.update( {
	"qld_on_new_button_clicked"	: qld_on_new_button_clicked,
	"qld_on_edit_button_clicked"	: qld_on_edit_button_clicked,
	"qld_on_delete_button_clicked"	: qld_on_delete_button_clicked,
	"qld_on_default_button_clicked"	: qld_on_default_button_clicked,
	"qld_on_apply_button_clicked"	: qld_on_apply_button_clicked
})

	

# =====================================================================================================================

def gui_main_run():
	called(gui_main_run)

	# Grey-out the window, (and draw it) to keep the user happy. "Nice User, want a donut?"
	qld_busy()

	# Gtk Signal Handler Connection
	# -----------------------------
	# All the gtk signal handlers are now collected in the gtk_signal_handler dictionary. A call to
	# signal_autoconnect will wire them up to the gtk callbacks, and they will be "real" signal handlers now.
	# (You are a real boy now, pinochio!)
	main.glade_context.signal_autoconnect(main.gtk_signal_handlers)

	# Initial Display
	# ---------------
	init_queue_edit_or_die(main.dynamic_box_name)
	foomatic_init_overview()
	qld_build_test_menu()
	debug_print(conf.print_tests)
	qld_list_queues()

	# restore regular cursor
	qld_ready()

	# Main Loop
	# ---------
	# Now that everything is defined, intialized, and conected, program execution can be handed over to
	# gtk mainloop, and callbacks and state machines described in the rest of the program will take care
	# of the rest.
	gtk.mainloop()


# =====================================================================================================================

def startup_and_find_cmd():
	if os.geteuid() != 0:
		print _("You must run printconf-gui as root.")
		sys.exit(-1)

	if "-h" in main.args or "--help" in main.args:
		print main.usage
		sys.exit(0)

	main.dynamic_box_name = "local"
	for arg in main.args:
		if arg[:6] == "--box=":
			main.dynamic_box_name = arg[6:]

	# Look for commands
	cmd = None
	for arg in main.args:
		if arg[0:3] == "--X":
			if cmd:
				sys.stderr.write(_("Conflicting commands: %s and %s") % (cmd, arg[3:]))
				sys.exit(1)
			else:
				cmd = arg[3:]
	if not cmd:
		cmd = "gui"
	
	if main.cmd_handlers.has_key(cmd):
		main.cmd_handlers[cmd]()
	else:
		sys.stderr.write(_("No handler for command: ") + cmd)
		sys.exit(1)
	

main.cmd_handlers = {
	"gui"	: gui_main_run,
}

main.args = sys.argv[1:]

main.usage = """printconf-tui [--X<command>] [<options>]
 options are:
  -h, --help	show usage

  --box=box_name
		set the edit box name. defaults to "local"
  
 available commands:
  gui		run the normal gui mode. this is the default

"""

	
# =====================================================================================================================

# Prolog
# ------
#
# I would recommend, to anyone wanting to write their own printing system, that they invest their efforts into
# extending and supporting IPP (Internet Printing Protocol) based printing. Any serious attempt to "fix" printing
# subsystems which use RFC1179 based protocols is doomed to failure, as the protocol lacks discovery, flexible media
# specifiaction, authentication, etc., and will probaly tweak your sanity.
#
# But then again, I was crazy before I started ;>
#				- Crutcher


